#!/bin/sh
# the next line restarts using wish \
exec wish "$0" -- "$@"

# Copyright 2001 Bryan Schofield
# This file is part of RoboTournament.
# 
# RoboTournament is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RoboTournament is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RoboTournament; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# Hide the root window
wm withdraw .

set cwd [pwd]
set script [info script]
while {[file type $script] == "link"} {
    cd [file dirname $script]
    set script [file readlink $script]
}
cd [file dirname $script]
set homeDir [pwd]
cd $cwd
set libDir [file join $homeDir lib]
lappend auto_path $libDir
unset cwd



set version 11.13.01


#=========================================================================
# Get any required packages
package require RTSSocket
package require RTSBot
package require RTSInsult
package require RTSError



#
# Set some defaults for look and feel
#
font create fnt-small \
    -family "Courier" \
    -size "12"
font create fnt-big \
    -family "Courier" \
    -size "16"

set background "\#004000"
set darkBackground "\#003000"
set highlight "\#50ff00"
set foreground "\#c0ffb0"
set darkForeground "\#508f40"
set contrastForeground "\#ffa000"
option add *Background $background
option add *HighlightBackground $background
option add *HighlightForeground $highlight
option add *Foreground $foreground 
option add *Relief flat
option add *Font "fnt-small"
option add *Button.HighlightThickness 1
option add *Button.HighlightBackground "black"
option add *Button.relief "flat"
option add *Button.Borderwidth 10
option add *Entry.HighlightThickness 1
option add *Entry.HighlightBackground "black"
option add *Radiobutton.HighlightThickness 1
option add *Radiobutton.HighlightBackground "black"
option add *Checkbutton.HighlightThickness 1
option add *Checkbutton.HighlightBackground "black"






#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                        Global Procs                          ======
#=====                                                              ======
#=========================================================================
#=========================================================================



#=========================================================================
# Global proc: pause --
#
# Pause by blocking while waiting for standard input.
# Useful for testing.
#
proc pause {} {
    puts "Paused: $RTS::game(turn).$RTS::game(phase) Enter to continue"
    gets stdin
}


#=========================================================================
# Global proc: bgerror --
#
# A quick hack for handling background errors that was needed
#
# proc bgerror { args } {
#     puts "RTS ERROR: [join $args]"
#     puts "$::errorCode"
#     puts "$::errorInfo"
#     rtssocket close
#     exit
# }


#=========================================================================
# Global proc: Usage --
#
# msg - optional message to print
#
#
# Prints usage info and exists
#
proc Usage { {msg ""} } {
   puts "RoboTournament Server"
   puts "Usage: $::script \[options\]"
   puts "Options:"
   puts "\t-help          (Display this help and exit)"
   puts "\t-port <number> (Port number)"
   puts "\n$msg"
   exit
}









#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                         RTS Namespace                        ======
#=====                                                              ======
#=========================================================================
#=========================================================================



namespace eval RTS {
    variable port 4390
    variable phaseDelay 1000     ;# The pause between phases
    variable robotDelay 250      ;# The pause between robots moving
    variable gameState "NeedMap" ;# General states of the server (game)
    # GAME STATES:
    #    NeedMap
    #    WaitingForMap
    #    WaitingToStart
    #    InProgress
    #    GameOver
    variable client              ;# Array of client data
    # client keys:
    #    client($id) - is set when client # $id exists
    #    client($id,botId) - id of the bot that this client controls
    #    client($id,bot) - bot widget that this client controls
    variable map                 ;# Huge array of map data, contains 
                                  # configuration of items
    variable item                ;# Huge array that contains details of items
    variable itemId 0            ;# Unique item id that is increment every
                                  # time a new item is created
    variable game                ;# Game data array
    set game(turn) 0             ;# The current turn
    set game(phase) 0            ;# The current phase
    set game(type) dm            ;# The game type
    set game(dmfrags) 5          ;# Number of frags required to win in DM games
    set game(rlyorder) [list m b g y p o] ; # Rally flag order
    set game(ctfcaptures) 5      ;# Number of capture to win CTF game
    variable team                ;# Array of team data
    set team(m) "Red"            ;# Name of the "m" team
    set team(b) "Blue"           ;# Name of the "b" team
    set team(m,members) ""       ;# Members of the "m" team (bot names)
    set team(b,members) ""       ;# Members of the "b" team (bot names)
    set team(m,score) 0          ;# Score of the "m" team
    set team(b,score) 0          ;# Score of the "b" team
}




#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                   List and Utility Procs                     ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::RandomizeList <alist>
#
# alist - any ol' list
#
#
# Returns a new list that is a randomized version of "alist"
#
proc RTS::RandomizeList { alist } {
    # puts "Randomize: $alist"
    set newList ""
    while {[llength $alist] > 0} {
        set index [expr {int(rand() * [llength $alist])}]
        lappend newList [lindex $alist $index]
        set alist [lreplace $alist $index $index]
    }
    #puts "       To: $newList"
    return $newList
}

#=========================================================================
# RTS::RemoveFromList <thing> <alist>
#
# thing - thing to remove
# alist - any ol' list
#
#
# Returns a new list created by removing all occurance of "thing" in
# "alist"
#
proc RTS::RemoveFromList { thing alist } {
    set newList ""
    foreach item $alist {
        if {$item != $thing} {
            lappend newList $item 
        }
    }
    return $newList
}




#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                   Init and Shut Down Procs                   ======
#=====                                                              ======
#=========================================================================
#=========================================================================

#=========================================================================
# RTS::Init <port>
#
# port - port number to use
#
#
# Init proc for starting up the server
#
proc RTS::Init { port } {
    set RTS::port $port
    rtsbot init 8
    RTS::Gui .rts
    # Better add anything you want to do during Init before
    # the rtssocket is opened...
    if {![rtssocket open $RTS::port]} {
        puts "Failed to open server socket on port $RTS::port"
    }    
}

#=========================================================================
# RTS::Gui <w>
#
# w - widget path of the tolevel to create
#
#
# Creates a simple window that displays some info about the server
# and provides a "Shut Down" button
#
proc RTS::Gui { w } {
   destroy $w
   toplevel $w
   wm title $w "RoboTournament Server"
   wm protocol $w WM_DELETE_WINDOW "RTS::ShutDown"
   wm geometry $w +0-0
   wm iconify $w
   set atts [list \
               "Turn" RTS::game(turn) \
               "Phase" RTS::game(phase) \
               "Game Type" RTS::game(type) \
               "Phase Delay" RTS::phaseDelay \
               "Robot Delay" RTS::robotDelay \
               "Port" RTS::port \
               "Game State" RTS::gameState \
               "Connections" RTSSocket::connections]
   set f [frame $w.atts]
   set i 0
   foreach {lab var} $atts {
      label $f.l$i \
         -text "$lab" \
         -anchor w
      label $f.e$i \
         -textvariable "$var" \
         -background "$::darkBackground" \
         -anchor w\
         -highlightbackground "black" \
         -highlightthickness 1
      grid $f.l$i $f.e$i -stick ew -padx 4 -pady 4
      incr i
   }

   button $w.shutdown \
      -text "Shut Down" \
      -command "RTS::ShutDown"\
         -background "$::darkBackground" \
         -highlightbackground "black" \
         -highlightthickness 2
      
   grid $w.atts -padx 4 -pady 4
   grid $w.shutdown -padx 4 -pady 4
      
}


#=========================================================================
# RTS::ShutDown --
#
#
# Simply closes the server socket and exits
#
proc RTS::ShutDown {} {
    rtssocket close
    exit
}









#=========================================================================
#=========================================================================
#=====                                                              ======
#=====              Message and Client Connection Handling          ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::MessageHandler <clientId> <message>
#
# clientId - id of the client from which the message originated
# message - the message
#
#
# Handles incoming message from the clients
# Messages:
# chat <text>
# register <robot_name> <robot_number>
# map <map_message_data> (See RTS::ProcessMap)
# program  <program_data> (See RTS::ProcessProgram)
# powerup <boolean>
# start --
#
proc RTS::MessageHandler {clientId message} {
    set type [lindex [split [join $message]] 0]
    set message [join [lrange [split [join $message]] 1 end]]
    switch -- [string tolower $type] {
        chat {
            rtssocket broadcast \
                "CHAT <[$RTS::client($clientId,bot) name]> $message"
        }
        register {
            set name [lindex [split $message] 0]
            set bot  [lindex [split $message] 1]
            if {![rtsbot isvalid $bot]} { set bot 0 }
            set ok 0
            if {[bot$bot free]} {
                set ok 1 
            } else {
                for { set bot 0 } {$bot < 8} { incr bot } {
                    if {[bot$bot free]} { 
                        set ok 1
                        break
                    }
                }
            }
            if {$ok} {
                bot$bot free 0
                bot$bot name $name
                bot$bot client $clientId
                RTS::NewClient $clientId $bot
                rtsbot sendnames
                return
            } else {
                rtssocket send $clientId "SERVERFULL Sorry, server is full"
                rtssocket disconnect $clientID
            }

        }
        map { RTS::ProcessMap $message $clientId }
        program { RTS::ProcessProgram $RTS::client($clientId,bot) $message }
        powerup { $RTS::client($clientId,bot) powerup $message }
        start {
            array unset RTS::game
            variable game
            set RTS::game(turn) 0
            set RTS::game(phase) 0
            set game(dmfrags) 5
            set game(rlyorder) [list m b g y p o]
            set game(ctfcaptures) 5

            set RTS::gameState "InProgress"
            switch -- $RTS::game(type) {
                rly {
                    foreach bot [rtsbot bots] {
                        $bot nextflag -1
                        RTS::AdvanceBotFlag $bot
                    }
                }
                ctf {
                    set numBots [llength [rtsbot bots]]
                    for {set i 0} {$i < [expr {$numBots/2}]} {incr i} {
                        [lindex [rtsbot bots] $i] team "m"
                    }
                    for {} {$i < $numBots} { incr i} {
                        [lindex [rtsbot bots] $i] team "b"
                    }
                    variable team
                    set team(m) "Red"
                    set team(b) "Blue"
                    set team(m,members) ""
                    set team(b,members) ""
                    set team(m,score) 0
                    set team(b,score) 0
                    foreach bot [rtsbot bots] {
                        lappend team([$bot team],members) [$bot name]
                        $bot nextflag ""
                        rtssocket broadcast "SETTEAM $bot [$bot team]"
                    }
                    rtssocket broadcast \
                        "SERVERMSG $team(m) Team: $team(m,members)"
                    rtssocket broadcast \
                        "SERVERMSG $team(b) Team: $team(b,members)"
                }
            }

            RTS::BeginTurn
        }
        default {
            if {[string length $type] > 0} {
               puts "Unkown Message RX: \"$type\" $message"
            }
        }
    }
}


#=========================================================================
# RTS::NewClient <clientId> <bot>
#
# clientId - the new client's unique id (typically socket handle)
# bot - robot number assigned to the client
#
#
# Completes the registration of the newly connected client
#
proc RTS::NewClient {clientId bot} {
    set RTS::client($clientId) 1
    set RTS::client($clientId,botId) $bot
    set RTS::client($clientId,bot) bot$bot
    # Let the client know who he is
    rtssocket send $clientId "YOUAREBOT $bot"
    # Let the everyone know he has joined the game
    rtssocket broadcast \
            "SERVERMSG [$RTS::client($clientId,bot) name] has joined"
    # Check the state, we may need to do something special
    switch -- $RTS::gameState {
        NeedMap {
            set RTS::gameState "WaitingForMap"
            rtssocket send $clientId "NeedMap"
        }
        WaitingForMap {
        }
        WaitingToStart {
            rtssocket send $clientId "MAP $RTS::map(data)"
        }
        InProgress {
            rtssocket send $clientId "MAP $RTS::map(data)"
            rtsbot sendupdate
        }
        GameOver {
        }
    }
}

#=========================================================================
# RTS::LostClient <clientId>
#
# clientId - unique client id
#
#
# This proc is called every time a registered client is lost/dropped
#
proc RTS::LostClient {clientId} {
    if {[info exists RTS::client($clientId)]} {
        rtssocket broadcast \
            "SERVERMSG [$RTS::client($clientId,bot) name] has dropped"
        rtssocket broadcast "NAME $RTS::client($clientId,bot) Dropped"
        $RTS::client($clientId,bot) free 1
        $RTS::client($clientId,bot) name "Dropped"
        $RTS::client($clientId,bot) client ""
        unset RTS::client($clientId)
        unset RTS::client($clientId,bot)
        unset RTS::client($clientId,botId)
    }
    #
    # See if we should automatically shut down
    #
    if {$RTSSocket::connections == 0} {
        RTS::ShutDown
    }
}





#=========================================================================
# RTS::ProcessProgram <bot> <program>
#
# bot - robot widget
# program - program data
#
# Program Data Format:
# <power_on_next> <weapon_number> <card0> <card1> ... <card4>
#
# Processes program data for bots, if every bot is programs, 
# calls to run the turn
#
proc RTS::ProcessProgram {bot program}  {
    # Set the bots program
    $bot poweronnext [lindex $program 0]
    $bot weapon [lindex $program 1]
    $bot program [lrange $program 2 end]
    #
    # Now check to see that anyone has not sent a program
    #
    set need ""
    foreach obot [rtsbot bots] {
        if {[llength [$obot program]] < 5} { 
            lappend need $obot
        }
    }
    #
    # If only one robot is not ready, send a message so everyone else
    # knows who is holding them up.
    # If everyone is good, procede with the turn
    #
    if {[llength $need] == 1} {
        rtssocket broadcast \
            "SERVERMSG Waiting on [$need name] to program its registers"
    } elseif {$need == ""} {
        RTS::RunTurn
    }
}


#=========================================================================
# RTS::ProcessMap <mapData> <clientId>
#
# mapData - map data .. duh.
# clientId - id of the client where the map data originated
#
#
# Map Data Format:
# <match_type> <card_freq_m1> m2> m3> <card_freq_jp> <card_freq_bu>
# <card_freq_tl> <card_freq_tr> <card_freq_ut> <card_freq_no> <card_freq_w1>
# <card_freq_w2> <card_freq_w3> <card_freq_w4> <card_freq_w5> <x> <y>
# <tile_data> <x> <y> <tile_data>  <x> <y> <tile_data> ...
#
# Sets game type, card frequency and map data for the game
# For info on tile data, see RTS::NewMapItem
#
proc RTS::ProcessMap { mapData clientId} {
    set RTS::gameState "WaitingToStart"
    variable map
    variable item
    variable itemId
    
    #
    # Reset the item id, clean the map and item arrays
    #
    set itemId 0
    array unset map
    array unset item
    
    #
    # Set the game type
    #
    set RTS::game(type) [lindex $mapData 0]
    
    #
    # Set the card frequency
    #
    set i 1
    set cards [list m1 m2 m3 jp bu tl tr ut no w1 w2 w3 w4 w5]
    foreach card $cards {
        rtsbot cardfrequency $card [lindex $mapData $i]
        incr i
    }

    #
    # (Re)Initialize some map data elements to nothing.
    # The following are lists of certain map items. 
    # Many procs require the existance of these variables
    # but they do not need has any items in their lists
    #
    set map(data) ""
    set map(lasers) ""
    set map(starts) ""
    set map(teleporters) ""
    set map(repairs) ""
    set map(flags) ""
    set map(floor) ""
    set map(bases) ""
    set map(navpoints) ""
    foreach flag $RTS::game(rlyorder) {
        set map(flag,$flag) ""
    }
    for { set n 0 } {$n < 99} { incr n } {
        if {[info exists map(navpoint,$n)]} {
            unset map(navpoint,$n)
        }
    }
    
    #
    # Parse the tile coordinate pair and data to create new items
    #
    foreach {x y tile} [split [lrange $mapData $i end]] {
        if {![string is integer $x] || ![string is integer $y]} {
            continue
        }
        RTS::NewMapItem $x $y $tile
    }

    #
    # Sort the nav points
    #
    set navs ""
    foreach itemId $map(navpoints) {
        set navId $item($itemId,id)
        set nav($navId) $itemId
        lappend navs $navId
    }
    set navs [lsort -integer $navs]
    set map(navpoints) ""
    foreach navId $navs {
        lappend map(navpoints) $nav($navId)
    }

    #
    # Make each robot an item of type "b"
    #
    foreach bot [rtsbot bots] {
        set item($bot) "b"
    }
    
    #
    # Broadcast the map data and notify the supplying client
    # that it is ok to start whenever
    #
    rtssocket broadcast "MAP $map(data)"
    rtssocket send $clientId "WANTTOSTART"
}
















#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                Turn and Phase Running Procs                  ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::BeginTurn --
#
#
# Begins a turn
#
proc RTS::BeginTurn {} {
   incr RTS::game(turn)
   # Let everyone know  a turn is beginning
   rtssocket broadcast "SERVERMSG Beginning turn $RTS::game(turn)"
   # Clean the programs
   foreach bot [rtsbot bots] {
       $bot program ""
   }
   # If any of the bots were killed, find a new place for them
   foreach bot [rtsbot bots] {
       if {![$bot placed]} { RTS::PlaceBot $bot }
   }
   # Let everyone know where robots are
   rtsbot sendupdate
   # Deal the cards, update the weapons
   set botWhoNeedToChoose ""
   foreach bot [rtsbot bots] {
       $bot dealcards
       $bot updateweapons
       if {[$bot free]} {
           # For computer controlled robots, have them pick there program
           lappend botWhoNeedToChoose $bot
       }
   }
   # We waited to have bots choose their programs until after we dealt
   # all of the cards. Bot AI can be cpu intensive, and this keeps
   # humans from waiting so much :)
   foreach bot $botWhoNeedToChoose {
      $bot chooseprogram
   }
   # With the programs dealt, the game will wait until all have sent 
   # valid programs, see the ProcessProgram proc
}
#=========================================================================
# RTS::RunTurn --
#
#
# Runs a turn
#
proc RTS::RunTurn {} {
    
    flush stdout
    set packet "RUNTURN"
    
    
    #
    # We need to check the bots, if they are powered down
    # reset their health to 100, and if they are using a
    # special weapon, let everyone else know it
    #
    foreach bot [rtsbot bots] {
        if {[$bot powereddown]} {
            append packet " $bot"
            if {[$bot health] < 100} { $bot health 100 }
        }        
        if {[$bot weapon] != 0} {
            rtssocket broadcast \
                "SERVERMSG [$bot name] is using $RTSBot::weaponName([$bot weapon])"
        }
    }
    # Broadcast the "RunTurn message with a list of powered down bots
    rtssocket broadcast $packet
    
    #
    # Run 5 phases, let everyone know when it is over, and begin a new turn
    #
    for {set i 0} {$i < 5} { incr i} {
        # puts "RUNTURN $RTS::game(turn).$i"
        RTS::RunPhase $i
        rtsbot sendupdate
        after $RTS::phaseDelay "set go 1"
        vwait go
        if {$RTS::gameState == "GameOver"} { return }
        rtssocket updateclients
        
    }
    rtssocket broadcast "ENDTURN"
    #
    # Also send the score if the game is CTF
    #
    if {$RTS::game(type) == "ctf"} {
        rtssocket broadcast "SERVERMSG *** Score: Red: $RTS::team(m,score)  -  Blue: $RTS::team(b,score)"
    }
    RTS::BeginTurn
}

#=========================================================================
# RTS::RunPhase <phase> [clone]
#
# phase - the phase number
# clone - Optional clone id
#
#
# Activates the board element in sequence.
# If clone is specified to be non-zero, then this run of a phase
# does not really count, esentially, it is being used for computer
# controlled robots to decide there program
#
proc RTS::RunPhase { phase { clone 0 } } {
   set RTS::game(phase) $phase
   # if {$clone == 0} {
   #    pause
   # }
   
   RTS::ActivateRobots $clone
   #puts "RTS::ActivateRobots [time {RTS::ActivateRobots $clone}]"
   RTS::ActivatePushers $clone
   #puts "RTS::ActivatePushers [time {RTS::ActivatePushers $clone}]"
   RTS::ActivateConveyors "d" $clone
   #puts "RTS::ActivateConveyors [time {RTS::ActivateConveyors d $clone}]"
   RTS::ActivateConveyors "c" $clone
   #puts "RTS::ActivateConveyors [time {RTS::ActivateConveyors c $clone}]"
   RTS::ActivateConveyors "d" $clone
   #puts "RTS::ActivateConveyors [time {RTS::ActivateConveyors d $clone}]"
   RTS::ActivateGears $clone
   #puts "RTS::ActivateGears [time {RTS::ActivateGears $clone}]"
   if {$clone == 0} { 
      RTS::ActivateRandomizers $clone 
      RTS::ActivateLasers $clone
   }
   #puts "RTS::ActivateLasers [time {RTS::ActivateLasers $clone}]"
   RTS::ActivateCrushers $clone
   #puts "RTS::ActivateCrushers [time {RTS::ActivateCrushers $clone}]"
   switch -- $RTS::game(type) {
       rly { 
           RTS::ActivateFlags $clone 
       }
       ctf { 
           RTS::ActivateFlags $clone 
           RTS::ActivateBases $clone 
       }
   }
   if {$phase == 4} {
      #puts "RTS::ActivateRepairs [time {RTS::ActivateRepairs $clone}]"
      RTS::ActivateRepairs $clone
   }
}


#=========================================================================
# RTS::GameOver <message>
#
# message - text message to broadcast as part of the GAMEOVER message
#
#
# Sends a GAMEOVER message to clients and sets gameState to "GameOver"
#
proc RTS::GameOver { message } {
    rtssocket broadcast "GAMEOVER $message"
    set  RTS::gameState GameOver
}














#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                   Board Element Activation                   ======
#=====                                                              ======
#=========================================================================
#=========================================================================

#*************************************************************************
#
#                             *** NOTE ***
#
# For all RTS::Activation* procs, the optional argument "clone" is
# used only for computer controlled robots to help them build there
# program. If the clone argument is set to be non-zero, then that
# activation does not really count. Meaning it does not effect the
# game at all. In some cases, activation is handled slightly different
# when "clone" is used.
#
#*************************************************************************



#=========================================================================
# RTS::ActivateRobots [clone]
#
# clone - optional, see NOTE above
#
#
# Parse robot programs and moves robots accordingly
#
proc RTS::ActivateRobots { {clone 0} } {
    variable item
    variable map

    #
    # Build an ordered list of robots to activate.
    # If not using a clone, order the bots
    # based on card priority. 
    # If we are using a clone, only activate it
    #
    if {$clone == 0} {
        #
        # Shuffle the robots, then check their cards
        #
        foreach bot [RTS::RandomizeList [rtsbot bots]] {
            if {![$bot placed]} { continue }
            # Get the card this bot is using
            set card [lindex [split [join [$bot program]]] $RTS::game(phase)]
            # If the is type "p" then it means that the computer should
            # pick a any card from the bots dealt program cards to use
            # this turn. This allows computer controlled bots to sorta
            # cheat. Which is ok,because they need the help!
            if {$card == "p"} {
                set card [$bot pickcard]
                $bot program [lreplace [split [join [$bot program]]] \
                                  $RTS::game(phase) $RTS::game(phase) $card] 
            }
            # Note the this bot is using this card
            lappend uses($card) $bot $card
        }
        #
        # Now, sort the bots based on following list of card priorities
        #
        foreach card [list m3 jp m2 m1 bu ut tr tl w1 w2 w3 w4 w5] {
            if {[info exists uses($card)]} {
                eval lappend orderedBots $uses($card)
            }
        }


        if {![info exists orderedBots]} { puts "No bots to move" ; return }




    } else {


        set orderedBots \
         [list $clone [lindex [$clone program] $RTS::game(phase)]]

    }

    #
    # Evaluate what card each bot is using in the predetermined order
    #
    foreach {bot card} $orderedBots {
        switch -- $card {
            bu { RTS::MoveBot $bot \
                     [RTS::GetRotatedDirection [$bot face] 180] 1 $clone}
            jp { RTS::JumpBot $bot [$bot face] $clone}
            m1 { RTS::MoveBot $bot [$bot face] 1 $clone}
            m2 { RTS::MoveBot $bot [$bot face] 2 $clone}
            m3 { RTS::MoveBot $bot [$bot face] 3 $clone}
            no { $bot takedamage -10 } 
            tr { $bot face [RTS::GetRotatedDirection [$bot face] 90]}
            tl { $bot face [RTS::GetRotatedDirection [$bot face] 270]}
            ut { $bot face [RTS::GetRotatedDirection [$bot face] 180]}
            w1 { $bot addweapon 1 }
            w2 { $bot addweapon 2 }
            w3 { $bot addweapon 3 }
            w4 { $bot addweapon 4 }
            w5 { $bot addweapon 5 }
            x  { }
            default {
                error "Unknown card: $card $bot phase: $RTS::game(phase) program: \"[$bot program]\"" 
            }
        }
        #
        # Check the bot position, if they are not dead already
        #
        if {[$bot placed]} {
            set x [$bot x]
            set y [$bot y]
            #
            # If for some reason a bot has moved into space, 
            # create a pit under him so he can die like he's 
            # suppose to
            #
            if {![info exists map($x,$y)]} {
                RTS::NewMapItem $x $y p01234
            }
            #
            # Check the items that bot is setting on
            # 
            foreach thing $map($x,$y) {
                set type $item($thing)
                switch -- $type {                    
                    p {
                        if {[regexp $RTS::game(phase) $item($thing,phases)]} {
                            #puts "Kill $bot - RTS::ActivateRobots"
                            $bot die
                            if {$clone == 0} {
                                rtssocket broadcast "FELLINPIT $bot"
                                rtssocket broadcast \
                                    "SERVERMSG [$bot name] fell in to a pit"
                            }
                        }
                    }
                }
            }
        } ; # End if {[$bot placed]}

        #
        # If we are not using clone, send an update
        #
        if {$clone == 0} {

            rtsbot sendupdate
            rtssocket updateclients
            after $RTS::robotDelay "set go 1"
            vwait go


        }
    }
}
#=========================================================================
# RTS::ActivatePushers [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "pusher" board elements
#
proc RTS::ActivatePushers { {clone 0} } {
    variable map
    variable item
    #
    # Create a list of bots either sorted randomly or just
    # using the optional clone
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
    } else {
        set bots $clone
    }
    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,u)]
            && [regexp $RTS::game(phase) $item($map($x,$y,u),phases)]} {
            RTS::MoveBot $bot $item($map($x,$y,u),dir) 1 $clone "pushed"
        }
    }
}
#=========================================================================
# RTS::ActivateRepairs [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "repair station" board elements
#
proc RTS::ActivateRepairs { {clone 0} } {
    variable map
    variable item
    #
    # Create a list of bots either sorted randomly or just
    # using the optional clone
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
    } else {
        set bots $clone
    }
    foreach bot $bots {
       if {![$bot placed]} { continue }
       set x [$bot x]
       set y [$bot y]
       if {[info exists map($x,$y,r)]} {
         $bot takedamage [expr -10 * {$item($map($x,$y,r),amount)}]
       }
    }
}

#=========================================================================
# RTS::ActivateBases [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "flag base" board elements
# Bases should only be on CTF games
#
proc RTS::ActivateBases { {clone 0} } {
    variable map
    variable item

    #
    # Clones do not get to do this. Neither do non-CTF matches
    #
    if {$RTS::game(type) != "ctf"} { return }
    if {$clone != 0} { return }

    set bots [RTS::RandomizeList [rtsbot bots]]

    foreach bot $bots {
        if {![$bot placed]} { continue }
        if {[join [$bot nextflag]] == ""} { continue }
        set x [$bot x]
        set y [$bot y]
        set base 0
        set flag 0
        #
        # Determine if the the flag and the base are both held
        # by this bots team
        #
        if {[info exists map($x,$y,e)] 
            && ($item($map($x,$y,e),team) == [$bot team])} {
                 set base 1
        }
        if {[info exists map($x,$y,a)] 
            && ($item($map($x,$y,a),team) == [$bot team])} {
                 set flag 1
        }
        #
        # If both flag and base are good, increment the score for
        # this bots team and re-place the flag
        #
        if {$base && $flag} {
            incr RTS::team([$bot team],score)
            # Let everyone there was a score
            rtssocket broadcast "SERVERMSG [$bot name] just scored a point for $RTS::team([$bot team]) Team"
            # Check the score
            if {$RTS::team([$bot team],score) >= $RTS::game(ctfcaptures)} {
                RTS::GameOver "$RTS::team([$bot team]) wins with $RTS::team([$bot team],score) captures"                
            }
            # Re-place the flag
            foreach baseId $RTS::map(bases) {
                if {[join [$bot nextflag]] == $RTS::item($baseId,team)} {
                    set x $RTS::item($baseId,x)
                    set y $RTS::item($baseId,y)
                }
            }
            RTS::PlaceItem [lindex $RTS::map(flag,[join [$bot nextflag]]) 0] \
                $x $y
            $bot nextflag ""
            rtssocket broadcast "SETFLAG $bot empty"
       }
    }
}

#=========================================================================
# RTS::ActivateFlags [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "flag" board elements
# Flag should only be in Rally and CTF matches
#
proc RTS::ActivateFlags { {clone 0} } {
    variable map
    variable item
    #
    # DM type matches and clones do not get to do this
    #
    if {$RTS::game(type) == "dm"} { return }
    if {$clone != 0} { return }
    
    set bots [RTS::RandomizeList [rtsbot bots]]

    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,a)]} {
           RTS::TouchFlag $bot $map($x,$y,a)
           # As a bonus, if this is the last phase
           # Give the bot a health bonus
           if {$RTS::game(phase) == 4} {
               $bot takedamage -10
           }
        }
    }
}
#=========================================================================
# RTS::ActivateConveyors <key> [clone]
#
# key - conveyor type "c" or "d" (Single or Double)
# clone - optional, see NOTE above
#
#
# Activates the "conveyor" board elements,both single and double
#
proc RTS::ActivateConveyors { key {clone 0} } {
    variable map
    variable item
    #
    # Create a list of bots either sorted randomly or just
    # using the optional clone
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
    } else {
        set bots $clone
    }
    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,$key)]} {
           RTS::MoveBot $bot $item($map($x,$y,$key),head) 1 $clone "pushed"
        }
    }
}

#=========================================================================
# RTS::ActivateGears [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "gear" board elements
#
proc RTS::ActivateGears { {clone 0} } {
    variable map
    variable item
    #
    # Create a list of bots either sorted randomly or just
    # using the optional clone
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
    } else {
        set bots $clone
    }
    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,g)]} {
           $bot face [RTS::GetRotatedDirection \
                               [$bot face] $item($map($x,$y,g),angle)]
        }
    }
}
#=========================================================================
# RTS::ActivateCrushers [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "crusher" board elements
#
proc RTS::ActivateCrushers { {clone 0} } {
    variable map
    variable item
    #
    # Create a list of bots either sorted randomly or just
    # using the optional clone
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
    } else {
        set bots $clone
    }
    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,h)]
            && [regexp $RTS::game(phase) $item($map($x,$y,h),phases)]} {
            $bot die
            if {$clone == 0} {
                rtssocket broadcast "CRUSHED $bot"
                rtssocket broadcast "SERVERMSG [$bot name] was crushed"
            }
        }
    }
}
#=========================================================================
# RTS::ActivateRandomizers [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "program randomizer" board elements
#
proc RTS::ActivateRandomizers { {clone 0} } {
    variable map
    variable item

    #
    # Clone do not get to do this
    #
    if {$clone != 0} { return }
    
    set bots [RTS::RandomizeList [rtsbot bots]]

    foreach bot $bots {
        if {![$bot placed]} { continue }
        set x [$bot x]
        set y [$bot y]
        if {[info exists map($x,$y,z)]
            && [regexp $RTS::game(phase) $item($map($x,$y,z),phases)]} {
            $bot program [RTS::RandomizeList [$bot program]]
            $bot safesend "RANDOMIZE  [join [$bot program]]"
        }
    }
}
#=========================================================================
# RTS::ActivateLasers [clone]
#
# clone - optional, see NOTE above
#
#
# Activates the "laser" board elements
#
proc RTS::ActivateLasers { {clone 0} } {
    variable map
    variable item
    set packet "LASERS "
        
    #
    # clones do not get affected by bot lasers, nor get to shoot
    #
    if {$clone == 0} {
        set bots [RTS::RandomizeList [rtsbot bots]]
        foreach bot $bots {
            if {![$bot placed] || [$bot powereddown]} { continue }
            set card [lindex [split [join [$bot program]]] $RTS::game(phase)]
            #
            # If the bot is currently playing a weapon upgrade or no-op
            # card, it does not get to shoot
            #
            if {[regexp {(^w[0-9]|no)} $card]} { continue }
            switch -- [$bot weapon] {
                0 {
                    RTS::FireLaser [$bot x] [$bot y] [$bot face] \
                        100 10 $bot "packet"
                }
                1 { 
                    RTS::FireLaser [$bot x] [$bot y] [$bot face] \
                        100 20 $bot "packet"
                }
                2 { 
                    RTS::FireLaser [$bot x] [$bot y] [$bot face] \
                        100 10 $bot "packet"
                    RTS::FireLaser [$bot x] [$bot y] \
                        [RTS::GetRotatedDirection [$bot face] 180] \
                        100 10 $bot "packet"
                }
                3 { 
                    RTS::FireLaser [$bot x] [$bot y] \
                        [RTS::GetRotatedDirection [$bot face] 90] \
                        100 10 $bot "packet"
                    RTS::FireLaser [$bot x] [$bot y] \
                        [RTS::GetRotatedDirection [$bot face] 270] \
                        100 10 $bot "packet"
                }
                4 {
                    RTS::FireLaser [$bot x] [$bot y] n 100 10 $bot "packet"
                    RTS::FireLaser [$bot x] [$bot y] e 100 10 $bot "packet"
                    RTS::FireLaser [$bot x] [$bot y] w 100 10 $bot "packet"
                    RTS::FireLaser [$bot x] [$bot y] s 100 10 $bot "packet"
                }
                5 {
                    RTS::FireLaser [$bot x] [$bot y] [$bot face] \
                        1 80 $bot "packet"
                }
                default {
                    puts "$bot has unknown weapon type [$bot weapon]"
                }
            }
        }
    }
    #
    # Fire the board lasers
    #
    foreach id $map(lasers) {
      RTS::FireLaser $item($id,x) $item($id,y) $item($id,dir) \
            $item($id,range) $item($id,damage) \
            "l$item($id,range)$item($id,damage)" "packet" $clone
      #set t [time {
      #  RTS::FireLaser $item($id,x) $item($id,y) $item($id,dir) \
      #      $item($id,range) $item($id,damage) \
      #      "l$item($id,range)$item($id,damage)" "packet" $clone}]
      #puts "RTS::FireLaser $t"
    }
    #
    # Broadcast the laser packet if not a clone
    #
    if {$clone == 0} {
        if {([llength $packet] > 1)} {
            rtssocket broadcast $packet
        }
    }
}

#=========================================================================
# RTS::FireLaser <x> <y> <dir> <range> <strength> <src> <packetVar> [clone]
#
# x - starting x coordinate
# y - starting y coordinate
# dir - direction of fire
# range - range of the laser
# strength - strength of the laser
# src - source of the laser (board laser item or robot)
# packetVar - the name of the packet variable in the calling proc
#             this is upvar'ed and appended to create the laser fire packet
# clone - optional clone , see NOTE above
#
#
# Fires a laser based on the input parameters
#
proc RTS::FireLaser { x y dir range strength src packetVar {clone 0} } {
    upvar $packetVar packet
    variable map
    variable item
    # Store the source x and y, this is used when creating the packet
    set srcX $x
    set srcY $y
    # Check the range, if the provide range is not an integer, consider it
    # infinite
    if {![string is integer $range]} { set range 100 }
    
    while {$range > 0} {
        incr range -1
        incr x [RTS::GetDx $dir]
        incr y [RTS::GetDy $dir]
        #
        # If we have no map data for the coord pair, 
        # continue firing
        #
        if {![info exists map($x,$y)]} { continue }
        #
        # Check each item in the coordinate pair for bots and walls
        #
        if {[info exists map($x,$y,b)]} {
            set thing $map($x,$y,b)
            #
            # If in clone mode, do not shoot other bots
            #
            if {($clone != 0) && ($thing != $clone)} { continue }
            # Determine how much damage to apply
            set thingDmg [expr {$strength \
                              - round(rand() * ($strength/10.0)) \
                              + round(rand() * ($strength/10.0))}]
                              
            
            #
            # Apply the damage, did we kill it?
            #
            if {[$thing takedamage $thingDmg]} {
                set affect "hit"
            } else {
                set affect "killed"
            }
            #
            # If the source is a bot, give some frag credit
            # If we killed something, give whole point
            # If we just inflicted some pain, add partial credit
            # and determine "who" was the source for a nice
            # message
            #
            set insultDmg $thingDmg
            if {[regexp "^bot" $src]} {
               if {[$thing placed]} {
                   $src addfrag [expr {$thingDmg / 100.0}]
               } else {
                   $src addfrag 1
                   set insultDmg 100
               }
               set who "[$src name]"
               if {$clone == 0} {
                  rtsinsult insult $src $insultDmg $thing
               }
            } else {
                set who "Board laser"
            }
            if {$clone == 0} {
               rtssocket broadcast "SERVERMSG $who $affect [$thing name], inflicting $thingDmg points of damage"
               append packet "$src $srcX $srcY $x $y "
            }
            return 
        } elseif {[info exists map($x,$y,w)]} {
            set thing $map($x,$y,w)
            # Check to see if the wall is active, if it is
            # the laser stops here
            if {[regexp $RTS::game(phase) $item($thing,phases)]} {
                return
            }
        }
    }
}













#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                    Movement Utility Procs                    ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::DirIsPerpindicular <dir> <baseDir>
#
# dir - test direction (heading)
# baseDir - base or comparison direction (heading)
#
#
# Returns boolean value if <dir> is 90 degrees (perpindicular)
# to the <baseDir> base direction
#
proc RTS::DirIsPerpindicular { dir baseDir} {
    #
    # Normalize dir and baseDir so that
    #  s == n
    #  w == e
    if {$dir == "s"} { set dir "n"}
    if {$dir == "w"} { set dir "e"}
    if {$baseDir == "s"} { set baseDir "n"}
    if {$baseDir == "w"} { set baseDir "e"}
    #
    # Simply compare the normalized directions, if they are equal
    # then they are not perpindicular
    #
    if {$dir == $baseDir} { 
      return 0
    } else { 
      return 1 
    }
}

#=========================================================================
# RTS::GetRotatedDirection <dir> <angle>
#
# dir - base direction (heading)
# angle - degrees in which to find the rotated base direction
#
#
# Find and return the direction (heading) that is rotated the amount
# specified by the angle
#
# Coordinate System:
#
#         90 (n)
#           |
# (w) 180 --+-- 0 (e)
#           |
#        270 (s)
#
#
proc RTS::GetRotatedDirection { dir angle } {
    switch -- $angle {
        90 {
            switch -- $dir {
                n {return e}
                e {return s}
                s {return w}
                w {return n}
            }
        }
        180 {
            switch -- $dir {
                n {return s}
                e {return w}
                s {return n}
                w {return e}
            }
        }
        270 {
            switch -- $dir {
                n {return w}
                e {return n}
                s {return e}
                w {return s}
            }
        }
    }
    return -code error "Bad angle ($angle) or direction ($dir)"
}
#=========================================================================
# RTS::GetDx <dir>
#
# dir - base direction (heading)
#
#
# return the delta or increment value for use when traversing map
# coordinates in the x direction
#
proc RTS::GetDx { dir } {
    switch -- $dir {
        n { set dx 0  }
        s { set dx 0  }
        e { set dx 1  }
        w { set dx -1 }
    }
    return $dx
}
#=========================================================================
# RTS::GetDx <dir>
#
# dir - base direction (heading)
#
#
# return the delta or increment value for use when traversing map
# coordinates in the y direction
#
proc RTS::GetDy { dir } {
    switch -- $dir {
        n { set dy -1 }
        s { set dy 1  }
        e { set dy 0  }
        w { set dy 0  }
    }
    return $dy
}













#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                        Robot Movement                        ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::JumpBot <bot> <dir> [clone]
#
# bot - the robot widget
# dir - direction (heading) to jump bot
# clone - optional clone, See NOTE above
#
#
# The jump bot proc. The jump algorithm is as follows:
# 1. Place the bot one square forward (by <dir>) without considering
#    what is on that square
# 2. Attemp to move bot one square forward (by <dir>)
#    If the move was success full, return, the has jumped
#    Otherwise procede
# 3. Place the bot on the square where he started in this jump proc
# 4. Move the bot one square forward (by <dir>), we don't care if
#    it is successful or not
#
# 
proc RTS::JumpBot {bot dir {clone 0}} {
    variable map
    variable item
    set saveX [$bot x]
    set saveY [$bot y]
    set x [expr {[$bot x] + [RTS::GetDx $dir]}]
    set y [expr {[$bot y] + [RTS::GetDy $dir]}]
    RTS::SetBotPlace $bot $x $y
    if {[RTS::MoveBot $bot $dir 1 $clone "pushed"]} { return }
    RTS::SetBotPlace $bot $saveX $saveY
    RTS::MoveBot $bot $dir 1 $clone "pushed"
}

#=========================================================================
# RTS::MoveBot <bot> <dir> <amount> [clone [pushed]]
#
# bot - robot widget
# dir - direction (heading) in which to move the bot
# amount - number square to move
# clone - optional clone, See NOTE above
# pushed - if clone is specified, "pushed" may also be
#          specified. A pushed robot can not push other robots
#
#
# This is the move bot routine, it is recursive and return a
# boolean to indicate if the move was successful
#
proc RTS::MoveBot {bot dir amount {clone 0} {pushed 0}} {
    variable map
    variable item



    #
    # Determine the new coords
    #
    set x [expr {[$bot x] + [RTS::GetDx $dir]}]
    set y [expr {[$bot y] + [RTS::GetDy $dir]}]
    #
    # If we have no data, this is off the edge of the map.
    # Put a pit there
    #
    if {![info exists map($x,$y)] || ([llength $map($x,$y)] == 0)} {
        RTS::NewMapItem $x $y p01234
    }
    #
    # Check all the items in the new coords
    # Make list of all the "things" there, and push the
    # any bots to the front of the list
    #
    set things ""
    foreach thing $map($x,$y) {
        if {[regexp "^bot" $thing]} {
            set things [linsert $things 0 $thing]
        } else {
            lappend things $thing
        }
    }


    #
    # Store any conveyor that we might find
    #
    set conveyor ""

    #
    # Interact with each thing
    #
    foreach thing $things {
        set type $item($thing)
        switch -- $type {
            c { set conveyor $thing }
            d { set conveyor $thing }
            b {
                if {$clone == 0} {
                    #
                    # bot is not a clone, try to push thing
                    #
                    if {$pushed == 0} {
                        #
                        # Try to push
                        #
                        if {[RTS::MoveBot $thing $dir 1 $clone pushed]} {
                            #
                            # Push was successful, did we push it to it's death
                            #
                            if {![$thing placed]} {
                                # Yes give some credit
                                rtsinsult insult $bot 100 $thing 
                                rtssocket broadcast "SERVERMSG [$bot name] pushed [$thing name] in a pit"
                                $bot addfrag 1
                            }
                            
                            
                        } else {
                            #
                            # Could not push, apply a little pain!
                            #
                            set thingDmg [expr {round(rand() * 5)}]
                            set botDmg [expr {round(rand() * 2)}]
                            $bot takedamage $botDmg
                            $thing takedamage $thingDmg
                            #
                            # Was it enough damage to kill the other robot?
                            #
                            if {[$thing placed]} {
                                #
                                # Did not kill it, but give some partial credit
                                #
                                $bot addfrag [expr {$thingDmg /  100.0}]
                                return 0
                            } else {
                                #
                                # The thing is dead, give some credit
                                #
                                rtssocket broadcast "SERVERMSG [$bot name] rammed [$thing name]"
                                $bot addfrag 1
                            }
                        }
                        
                        
                    } else {
                        #
                        # bot was pushed and can not move, return failure
                        #
                        return 0
                    }
                    
                    
                } ; # End if {$clone == 0}
            }
            w {
                if {[regexp $RTS::game(phase) $item($thing,phases)]} {
                    #
                    # The bot ran into a wall that is active this phase
                    # Apply a little damage and return failure
                    #
                    set botDmg [expr {round(rand() * 2)}]
                    $bot takedamage $botDmg
                    return 0
                }
            }
            p {
                if {[regexp $RTS::game(phase) $item($thing,phases)]} {
                    #
                    # The bot ran into a pit that is active this phase
                    # Die!
                    #
                    $bot die
                    #
                    # If this was not a clone, broad cast a message
                    # letting everyone know about it
                    #
                    if {$clone == 0} {
                        rtssocket broadcast "FELLINPIT $bot"
                        rtssocket broadcast \
                            "SERVERMSG [$bot name] fell in to a pit"
                    }
                    #
                    # The move was successful, though I doubt the robot 
                    # thought so!
                    #
                    return 1
                }
            }
            l { 
               #
               # The bot ran into a laser
               # Apply a little damage and return failure
               #
               set botDmg [expr {round(rand() * 2)}]
               $bot takedamage $botDmg
               return 0 
            }
            o {
               #
               # Keep on moving, increment the amount counter because we
               # hit oil!
               #
               incr amount 
              }
            t {
                #
                # Look for a match to the teleporter, however, if there
                # are less than 2 matching this group, don't bother
                #
                if {[llength $map(teleporters,$item($thing,group))] < 2} { 
                    continue 
                }
                #
                # Look at each teleporter in random order
                #
                foreach telep [RTS::RandomizeList \
                                   $map(teleporters,$item($thing,group))] {
                    #
                    # If this teleporter is the one the robot is on, 
                    # keep looking
                    #
                    if {$telep == $thing} { continue }
                    #
                    # Check to see if the teleporter is empty
                    #
                    set teleport 1
                    foreach otherThing $map($item($telep,x),$item($telep,y)) {
                        if {$item($otherThing) == "b"} { 
                            # Can not teleport here!
                            set teleport 0 
                            break 
                        }
                    }
                    #
                    # If we can teleport, do it
                    #
                    if {$teleport} {
                        RTS::SetBotPlace $bot $item($telep,x) $item($telep,y)
                        incr amount -1
                        if {$amount > 0} {
                            RTS::MoveBot $bot $dir $amount $clone
                        }    
                        return 1
                    }
                } ; #End the foreach telep
            }
        } ; # End switch -- $type 
    }; # End foreach thing $things

    #
    # If it is not ok to move, then we should have returned failure already
    #

    incr amount -1
    
    #
    # Check to see if we are now on a conveyor and need to be rotated
    # The criteria are:
    # We are now on a conveyor, 
    # We are not still moving, that is we are stopping on it
    # and it has a tail that is not 180 degrees from our
    # direction of travel
    #
    if {($conveyor != "") && ($amount == 0)
        && ([lsearch $item($conveyor,tails) \
                 [RTS::GetRotatedDirection $dir 180]] != -1)} {
        #
        # Now rotate the bot if the head of the conveyor
        # and robot dir of travel meet certain criteria
        #
        switch -- $item($conveyor,head) {
            n {
                switch -- $dir {
                    e {$bot face [RTS::GetRotatedDirection [$bot face] 270]}
                    w {$bot face [RTS::GetRotatedDirection [$bot face] 90]}
                }
            }
            e {
                switch -- $dir {
                    n {$bot face [RTS::GetRotatedDirection [$bot face] 90]}
                    s {$bot face [RTS::GetRotatedDirection [$bot face] 270]}
                }
            }
            s {
                switch -- $dir {
                    w {$bot face [RTS::GetRotatedDirection [$bot face] 270]}
                    e {$bot face [RTS::GetRotatedDirection [$bot face] 90]}
                }
            }
            w {
                switch -- $dir {
                    s {$bot face [RTS::GetRotatedDirection [$bot face] 90]}
                    n {$bot face [RTS::GetRotatedDirection [$bot face] 270]}
                    }
            }
        }                        
    } ; # End if {($conveyor != "")}
    #
    # Formally, place the bot and if we are still moving, move again
    #
    RTS::SetBotPlace $bot $x $y
    if {$amount > 0} {
        RTS::MoveBot $bot $dir $amount $clone
    }
    # Return success
    return 1
}














#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                       Robot Placement                        ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::RemoveBotFromMap <bot>
#
# bot - robot widget
#
#
# Removes the bot from the map. If for some reason the bot is found
# where it is suppose to be, we absolutely need to find it. So search.
#
proc RTS::RemoveBotFromMap { bot } {
    variable map
    set bx [$bot x]
    set by [$bot y]
    if {[info exists map($bx,$by)]} {
        if {[lsearch $map($bx,$by) $bot] == -1} {
            foreach name [array names map] {
                if {[lsearch $map($name) $bot] != -1} {
                    set map($name) [RTS::RemoveFromList $bot $map($name)]
                }
            }
        } else {
            set map($bx,$by) [RTS::RemoveFromList $bot $map($bx,$by)]
            set otherBotIndex [lsearch -regexp $map($bx,$by) "^b"]
            if {$otherBotIndex == -1} {
               unset map($bx,$by,b) 
            } else {
               set map($bx,$by,b) [lindex $map($bx,$by) $otherBotIndex]
            }
        }
    }
}

#=========================================================================
# RTS::SetBotPlace <bot> <x> <y> [face [force]]
#
# bot - bot widget
# x - the x coordinate to place the bot
# y - the y coordinate to place the bot
# face - optional direction (heading) to use when placing he bot
# force - if face is specified, you may also force the the bot
#         to be placed by setting the "force" argument to be non-zero
#         Bots are not, by default, placed if their "placed" attribute
#         if false. [$bot placed] indicates if the bot is currently
#         placed on the map
#
#
# Sets the bot place by updating the map and bot data.
#
proc RTS::SetBotPlace {bot x y {face ""} {force 0}} {
    variable map
    if {![$bot placed] && ($force == 0)} { return 0 }
    #
    # Remove the bot from the map
    #
    RTS::RemoveBotFromMap $bot
    #
    # Set face if specified
    #
    if {$face != ""} {
        $bot face $face
    }
    #
    # Set the bot x,y and placed attributes
    #
    $bot x $x
    $bot y $y
    $bot placed 1
    #
    # Append the map data to include the new bot location
    #
    lappend map($x,$y) $bot
    set map($x,$y,b) $bot
    #
    # The following code is to check when bots are being placed
    # on top of each other.
    #
    # set someBots ""
    # foreach thing $map($x,$y) {
    #     switch -- $RTS::item($thing) {
    #         b {
    #             if {![rtsbot isclone $thing]} {
    #                 lappend someBots $thing
    #             }
    #         }
    #     }
    # }
    # if {![rtsbot isclone $bot] && ([llength $someBots] > 1)} {
    #     regsub "^bot" $bot "" botId
    #     puts stderr "ERROR: Too many bots on $x,$y : $someBots"
    #     return 0
    # } else {
    #     return 1
    # }
    return 1
}

#=========================================================================
# RTS::PlaceBot <bot>
#
# bot - robot widget
#
#
# Places unplaced robots on the map on some start position
#
proc RTS::PlaceBot { bot } {
    variable map
    variable item
    #
    # First look at the starts in random order
    #
    foreach i [RTS::RandomizeList $map(starts)] {
        set occupied 0
        #
        # Check to see if there are any bots on that start
        # If so, then this space is occupied, and we should
        # look at the other starts
        #
        foreach thing $map($item($i,coords)) {
            if {[regexp "^bot*" $thing]} { 
                set occupied 1
                break
            }
        }
        if {!$occupied} {
            set place $i
            break
        }
    }
    #
    # Check to see if we found a place to out the bot.
    # If we did not, then there are insufficient places to start
    # We must resort to putting the robot on any floor space
    #
    if {![info exists place]} {
        puts stderr "Insufficient start positions. Now using all floor spaces"
        set pos [list n e s w]
        foreach itemId $RTS::map(floor) {
            if {[llength $RTS::map($RTS::item($itemId,coords))] == 1} {
               RTS::NewMapItem $RTS::item($itemId,x) $RTS::item($itemId,y) \
                  s[lindex [RTS::RandomizeList $pos] 0]
            }
        }
        #
        # Place the bot again.
        #
        RTS::PlaceBot $bot
        return
    }
    #
    # We have a place to be, so set it now... and use force
    #
    RTS::SetBotPlace $bot $item($place,x) \
              $item($place,y) $item($place,facing) "starting"
}













#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                        Flag Touching                         ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::TouchFlag <bot> <itemId>
#
# bot -  bot widget
# itemId - the item id of the flag item
#
# 
# Handles touching of flags for all game types
#
proc RTS::TouchFlag {bot itemId} {
    if {[rtsbot isclone $bot]} { return }
    switch -- $RTS::game(type) {
        rly {
            set neededFlag [lindex $RTS::game(rlyorder) [$bot nextflag]]
            if {$neededFlag == ""} { return }
            if {$itemId != $RTS::map(flag,$neededFlag)} { return }
            RTS::AdvanceBotFlag $bot
        }
        ctf {
            if {[$bot team] == $RTS::item($itemId,team)} {
                # return the flag
                RTS::ReturnFlag $bot $itemId
            } elseif {[join [$bot nextflag]] == ""} {
                $bot nextflag $RTS::item($itemId,team)
                puts "SERVERMSG [$bot name] grabs the flag"
                rtssocket broadcast "SERVERMSG [$bot name] grabs the flag"
                rtssocket broadcast "SETFLAG $bot $RTS::item($itemId,team)"
                RTS::RemoveItem $itemId
            }
        }
    }
}

#=========================================================================
# RTS::AdvanceBotFlag <bot>
#
# bot -  bot widget
#
#
# Advances the next needed flag for the bot. Skips any flags that are
# specified in RTS::game(rlyorder) that are not on the board
#
proc RTS::AdvanceBotFlag { bot } {
    set advancing 1
    while {$advancing} {
        $bot nextflag [expr {[$bot nextflag] + 1}]
        if {[$bot nextflag] >= [llength $RTS::game(rlyorder)]} {
            RTS::GameOver "[$bot name] has won"
            return
        }
        set flag [lindex $RTS::game(rlyorder) [$bot nextflag]]
        if {$RTS::map(flag,$flag) != ""} {
            rtssocket broadcast "SETFLAG $bot $flag"
            return
        }
    }
}

#=========================================================================
# RTS::ReturnFlag <bot> <itemId>
#
# bot -  bot widget
# itemId - the item id of the flag item
#
# 
# Handles returning the flag from the field to its base for CTF games
#
proc RTS::ReturnFlag { bot itemId } {
    variable map
    variable item
    foreach baseId $map(bases) {
        if {$item($itemId,team) == $item($baseId,team)} {
            if {$item($itemId,coords) == $item($baseId,coords)} { return }
            $bot nextflag ""
            rtssocket broadcast "SETFLAG $bot empty"
            rtssocket broadcast "SERVERMSG [$bot name] returns the flag"
            puts "RETURNFLAG $bot $itemId"
            RTS::PlaceItem $itemId $item($baseId,x) $item($baseId,y)
            return
        }
    }
}










#=========================================================================
#=========================================================================
#=====                                                              ======
#=====                  Item Creation and Placement                 ======
#=====                                                              ======
#=========================================================================
#=========================================================================


#=========================================================================
# RTS::NewMapItem <x> <y> <tile>
#
# x - x coordinate to place tile
# y - y coordinate to place tile
# tile - string that indicates the properties of a tile
#
#
# Creates a new map item and places it at x,y
#
# Tile Format: <tile_type><tile_spec>
# where tile spec is 1 or more letters
# TILES:
# Flag            : a<color_or_team>
# Bot             : b 
# Flag Base       : e<color_or_team>
# Conveyor        : c<head><tail>[tail...tail]
# Double Conveyor : d<head><tail>[tail...tail]
# Floor           : f
# Gear            : g<angle> (angle = 90,180,270)
# Crusher         : h<phases> (phases = combo of 0,1,2,3,4)
# Laser           : l<dir><range><damage> (dir=n,s,e,w range=0-9,i damage=0-9)
# Nav Point       : n<nav_point_id>
# Oil             : o
# Pit             : p<phases> (phases = combo of 0,1,2,3,4)
# Repair Station  : r<amount>
# Start Position  : s<facing> (facing = n,s,e,w)
# Teleporters     : t<group>
# Pushers         : u<dir><phases> (dir=n,s,e,w phases = combo of 0,1,2,3,4)
# Wall            : w<phases> (phases = combo of 0,1,2,3,4)
# Randomizer      : z<phases> (phases = combo of 0,1,2,3,4)
#
# NOTE: "b" SHOULD NEVER BE ISSUED WITH NewMapItem IT IS PLACEHOLDER
proc RTS::NewMapItem { x y tile } {
    variable item
    variable map
    variable itemId


    set tile [string tolower $tile]
    set type [string index $tile 0]
    set spec [string range $tile 1 end]
    
    switch -- $type {
        a {
            set map(flag,$spec) $itemId
            set item($itemId,team) $spec
        }
        b {} 
        e {
            set item($itemId,team) $spec
            lappend map(bases) $itemId
        }
        c {
            set item($itemId,head) [string index $spec 0]
            for {set j 1} {$j < [string length $spec]} {incr j} {
                lappend item($itemId,tails) [string index $spec $j]
            }
        }
        d {
            set item($itemId,head) [string index $spec 0]
            for {set j 1} {$j < [string length $spec]} {incr j} {
                lappend item($itemId,tails) [string index $spec $j]
            } 
        }
        f {
            lappend map(floor) $itemId
        }
        g {
            set item($itemId,angle) $spec
        }
        h {
            set item($itemId,phases) $spec
            }
        l {
            set item($itemId,dir) [string index $spec 0]
            set item($itemId,range) [string index $spec 1]
            set item($itemId,damage) [expr {[string index $spec 2] * 10}]
            lappend map(lasers) $itemId
        }
        n {
            set item($itemId,id) $spec
            lappend map(navpoints) $itemId
            set map(navpoint,$spec) $itemId
        }
        o {
        }
        p {
            set item($itemId,phases) $spec
        }
        r {
            set item($itemId,amount) $spec
            lappend map(repairs) $itemId
        }
        s {
            set item($itemId,facing) $spec
            lappend map(starts) $itemId
        }
        t {
            set item($itemId,group) $spec
            lappend map(teleporters,$spec) $itemId
        }
        u {
            set item($itemId,dir) [string index $spec 0]
            set item($itemId,phases) [string range $spec 1 end]
        }
        w {
            set item($itemId,phases) $spec
        }
        z {
            set item($itemId,phases) $spec
        }
        default { 
            puts "Unknown Map Item: $tile" 
            continue 
        }
    }


    set item($itemId) $type
    set item($itemId,coords) "$x,$y"
    set item($itemId,x) $x
    set item($itemId,y) $y
    lappend map($x,$y) $itemId
    set map($x,$y,$type) $itemId

    #
    # Append the map data for every type of item except for starts
    # and nav points. The map data is sent to the clients so they
    # can draw the map. They don't need either of these two things
    #
    if {($type != "s") && ($type != "n")} {
        set map(data) "$map(data)$x $y t$itemId $tile "
    }

    incr itemId
}

#=========================================================================
# RTS::RemoveItem <itemId>
#
# itemId - unique id of the item
#
#
# Remove the item from the map, but does not delete or destroy the item
#
proc RTS::RemoveItem {itemId} {
    variable map
    variable item
    set oldX $item($itemId,x)
    set oldY $item($itemId,y)
    set map($oldX,$oldY) [RTS::RemoveFromList $itemId $map($oldX,$oldY)]
    if {[info exists map($oldX,$oldY,$item($itemId))]} {
      unset map($oldX,$oldY,$item($itemId))
    }
    puts "REMOVEITEM $itemId"
    rtssocket broadcast "REMOVEITEM $itemId"
}


#=========================================================================
# RTS::MoveItem <itemId> <x> <y>
#
# itemId - unique id of the item
# <x> - x coordinate to place the item
# <y> - y coordinate to place the item
#
#
# Removes the item from it's current location and places it at the
# coordinates specified by x,y
#
proc RTS::MoveItem {itemId x y} {
    variable map
    variable item
    set oldX $item($itemId,x)
    set oldY $item($itemId,y)
    set map($oldX,$oldY) [RTS::RemoveFromList $itemId $map($oldX,$oldY)]
    set item($itemId,coords) "$x,$y"
    set item($itemId,x) $x
    set item($itemId,y) $y
    set map($x,$y,$item($itemId)) $itemId
    lappend map($x,$y) $itemId
    puts "MOVEITEM $itemId $x $y"
    rtssocket broadcast "MOVEITEM $itemId $x $y"
}


#=========================================================================
# RTS::PlaceItem <itemId> <x> <y>
#
# itemId - unique id of the item
# <x> - x coordinate to place the item
# <y> - y coordinate to place the item
#
#
# PLaces the item at the coordinates specified by x,y with out 
# attempting to remove it from it's current location
#
proc RTS::PlaceItem {itemId x y} {
    variable map
    variable item
    set item($itemId,coords) "$x,$y"
    set item($itemId,x) $x
    set item($itemId,y) $y
    lappend map($x,$y) $itemId
    puts "PLACEITEM $itemId $x $y - $item($itemId)"
    rtssocket broadcast "PLACEITEM $itemId $x $y"
}


#========================================================================
#========================================================================
#========================================================================
#========================================================================
#========================================================================
#========================================================================
#========================================================================


#
# Process the command line arguments and initialize
#
set port 4390
for {set i 0} {$i < [llength $argv]} {incr i} {
   set opt [lindex $argv $i]
   switch -glob -- [string tolower $opt] {
      -p* { incr i ; set port [lindex $argv $i] }
      -h* { Usage }
      default { Usage "Invalid argument: $opt" }
   }
}

RTS::Init $port
