# Copyright 2001 Bryan Schofield
# This file is part of RoboTournament.
# 
# RoboTournament is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# RoboTournament is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with RoboTournament; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package provide RTSSocket 1.0


proc ::rtssocket { subcommand args } {
    switch -- [string tolower $subcommand] {
        open {
            if {[llength $args] < 1} {
                return -code error \
                    "Invalid arguments: rtssocket open <port> \[-debug\]"
            } else {
                return [RTSSocket::Open [lindex $args 0] [lindex $args 1]]
            }
        }
        close {
            return [RTSSocket::Close]
        }
        broadcast {
            if {[llength $args] != 1} {
                 return -code error \
                     "Invalid arguments: rtssocket broadcast <message>"
            } else {
                return [RTSSocket::Broadcast $args]
            }
        }
        send {
            if {[llength $args] != 2} {
                 return -code error \
                     "Invalid arguments: rtssocket send <client> <message>"
            } else {
                return [RTSSocket::Send [lindex $args 0] [lindex $args 1]]
            }
        }
        updateclients {
            return [RTSSocket::UpdateClients]
        }
        default {
            return -code error \
                "Invalid arguments: Should be:\n\trtssocket open <port> \[-debug\]\n\trtssocket close\n\trtssocket send <message>\n\trtssocket broadcast <message>\n\trtssocket updateclients"
        }
    }
}


namespace eval RTSSocket {
    variable debug 0
    variable clients ""
    variable connections 0
    variable maxConnections 8
    variable serving 0
    variable serverID
    variable allClientsUpdated
    variable numClientsUpdated
}

#========================================================================
proc RTSSocket::Dputs {msg} {
    if {$RTSSocket::debug} {
        puts stderr "DEBUG> $msg"
    }
}

#========================================================================
proc RTSSocket::Open {port args} {
    if {$RTSSocket::serving} {
        return
    }
    switch -- $args {
        -debug {set RTSSocket::debug 1}
    }
    if {[catch {socket -server RTSSocket::Connect $port} id]} {
      tk_messageBox \
         -title "RoboTournament Server" \
         -icon error \
         -message "Failed to start server: $id"
      exit
    }
    set RTSSocket::serverID $id
    set RTSSocket::serving 1
    RTSSocket::Dputs "RTSSocket Started"
    vwait forever
}


#========================================================================
proc RTSSocket::Connect {client addr port} {
    if {[llength $RTSSocket::clients] 
        >= $RTSSocket::maxConnections} {
        close $client
        return
    }
    fileevent $client readable "RTSSocket::Receive $client"
    fconfigure $client -buffering line
    lappend RTSSocket::clients $client
    incr RTSSocket::connections
    RTSSocket::Dputs "NEW_CLIENT $client"
}

#========================================================================
proc RTSSocket::Receive {client} {
    if {[eof $client]} {
        RTSSocket::Disconnect $client
        return
    }
    gets $client message
    RTSSocket::Dputs "RTSSocket Received: $message"
    if {$message == "I_AM_UPDATED"} {
        RTSSocket::ClientUpdated $client
    } else {
        RTS::MessageHandler $client $message
    }
}


#========================================================================
proc RTSSocket::Send {client message} {
    RTSSocket::Dputs "RTSSocket Sending: \($client\): $message"
    if {[catch {puts $client $message } err]} {
        RTSSocket::Dputs "RTSSocket Send Error: $err"
    }
    catch { flush $client }
}

#========================================================================
proc RTSSocket::Broadcast {message} {
    foreach client $RTSSocket::clients {
        RTSSocket::Send $client $message
    }
}

#========================================================================
proc RTSSocket::Disconnect {client} {
    close $client
    incr RTSSocket::connections -1
    set list ""
    foreach c $RTSSocket::clients {
        if {$client != $c} {
            lappend list $c
        }
    }
    set RTSSocket::clients $list
    RTS::LostClient $client
    RTSSocket::Dputs "RTSSocket Disconnect: $client"
}

#========================================================================
proc RTSSocket::UpdateClients {} {
    set RTSSocket::allClientsUpdated 0
    set RTSSocket::numClientsUpdated 0
    RTSSocket::Broadcast "UPDATE_IDLE_TASKS"
    vwait RTSSocket::allClientsUpdated
}

#========================================================================
proc RTSSocket::ClientUpdated { client } {
    incr RTSSocket::numClientsUpdated
    RTSSocket::Dputs \
        "Number of Updated Clients: $RTSSocket::numClientsUpdated"
    if {$RTSSocket::numClientsUpdated 
        >= [llength $RTSSocket::clients]} {
        set RTSSocket::allClientsUpdated 1
        RTSSocket::Dputs "All Clients Updated"
    }
}

#========================================================================
proc RTSSocket::Close {} {
    if {$RTSSocket::serving} {
        foreach client $RTSSocket::clients {
            RTSSocket::Disconnect $client
        }
        close $RTSSocket::serverID
        set RTSSocket::serving 0
        RTSSocket::Dputs "Closed RTSSocket"
    }
}
